/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2013                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#ifndef SCOUT_CBDATA_H
#define SCOUT_CBDATA_H


#include <stack>

#include <pearl/CallbackData.h>
#include <pearl/Event.h>

#include "scout_types.h"
#include "fixed_set.hh"

/*-------------------------------------------------------------------------*/
/**
 * @file  CbData.h
 * @brief Declaration of the CbData struct.
 *
 * This header file provides the declaration of the CbData struct which
 * stores the data required for the pattern search replay.
 */
/*-------------------------------------------------------------------------*/


//--- Forward declarations --------------------------------------------------

namespace pearl
{

class LocalData;
class RemoteData;
class GlobalDefs;
class LocalTrace;

}   // namespace pearl


namespace scout
{
/*-------------------------------------------------------------------------*/

class SynchpointHandler;
class TmapCacheHandler;
class OmpEventHandler;

/**
 * @struct CbData
 * @brief  Datatype aggregating all data which is required for the pattern
 *         search replay.
 *
 * The CbData struct provides a datatype which aggregates a number of values
 * which are required during the pattern search replay. In most cases, this
 * is data which has to be passed between callbacks (i.e., from a base pattern
 * to a more specific sub-pattern).
 *
 * @todo Review, potentially rename and document all undocumented attributes.
 */
/*-------------------------------------------------------------------------*/

struct CbData
    : public pearl:: CallbackData
{
    /// @brief   Local data
    /// @details Collects local pattern constituents of point-to-point
    ///          communication patterns before sending them to the peer.
    pearl::LocalData* mLocal;

    /// @brief   Remote data
    /// @details Stores remote pattern constituents of point-to-point
    ///          communication patterns after receiving them from the peer.
    pearl::RemoteData* mRemote;

    /// @brief   Local event set
    /// @details Collects local pattern constituents of point-to-point
    ///          communication patterns before sending them to the peer.
    pearl::LocalData* mInvLocal;

    /// @brief   Remote event set
    /// @details Stores remote pattern constituents of point-to-point
    ///          communication patterns after receiving them from the peer.
    pearl::RemoteData* mInvRemote;

    /// Global definitions object
    pearl::GlobalDefs* mDefs;
    /// Local trace object
    pearl::LocalTrace* mTrace;

    /// MPI Synchpoint Handler

    SynchpointHandler*    mSynchpointHandler;
    TmapCacheHandler*     mTmapCacheHandler;
    OmpEventHandler*      mOmpEventHandler;

    /// Container for Information on collective operations
    CollectiveInfo mCollinfo;

    /// @brief   Detected waiting time
    /// @details Used to transfer the already calculated waiting time to
    ///          sub-patterns.
    pearl::timestamp_t mIdle;

    /// rank for synchpoints
    int mSyncRank;

    /// temporary storage for an event
    pearl::Event mEvent;

    /// time and local rank for RMA ops
    TimeRank mLastPost;
    TimeRank mLastComplete;
    TimeRank mLastRmaOp;

    TimeVec2 mTimevec;

    /// Stores GATS-Synchpoint ranks in backwards replay
    fixed_set<int> mGatsSyncRanks;

    /// Stores remote comm_id for WfP
    uint32_t mCommId;

    /// Stores extra info for WfP synchpoints
    CommInfoMap mSyncPointCommInfo;


    /// @brief Detected count;
    int mCount;

    /// @brief Number of Targets in RMA epoch
    int mCountTarget;

    /// @brief Number of Origins in RMA epoch
    int mCountOrigin;

    /// Current call stack
    std::stack<pearl::Event> mCallstack;
    bool                     mUpdateCallstack;

    /// @}
    /// @name Asynchronous analysis
    /// @}

    /// @brief Cnode id
    pearl::Callpath::IdType mCallpathId;

    /// window pointer
    pearl::MpiWindow* mWindow;

    std::set<int>* mLocations;

    /// @}
    /// @name Constructors & destructor
    /// @{

    CbData();
    virtual
    ~CbData();

    /// @}
    /// @name Pre- and postprocessing
    /// @{

    virtual void
    postprocess(const pearl::Event& event);

    /// @}
};
}   // namespace scout


#endif   // !SCOUT_CBDATA_H
