/*
 * This file is part of the Score-P software (http://www.score-p.org)
 *
 * Copyright (c) 2009-2013,
 * RWTH Aachen University, Germany
 *
 * Copyright (c) 2009-2013,
 * Gesellschaft fuer numerische Simulation mbH Braunschweig, Germany
 *
 * Copyright (c) 2009-2013,
 * Technische Universitaet Dresden, Germany
 *
 * Copyright (c) 2009-2013,
 * University of Oregon, Eugene, USA
 *
 * Copyright (c) 2009-2013,
 * Forschungszentrum Juelich GmbH, Germany
 *
 * Copyright (c) 2009-2013,
 * German Research School for Simulation Sciences GmbH, Juelich/Aachen, Germany
 *
 * Copyright (c) 2009-2013,
 * Technische Universitaet Muenchen, Germany
 *
 * This software may be modified and distributed under the terms of
 * a BSD-style license.  See the COPYING file in the package base
 * directory for details.
 *
 */

/** @internal
 *
 *  @file
 *
 *  @brief      This layer is completely internal to the OTF2 library and should
 *              not be exposed to the end user. Its purpose is to abstract the
 *              file interaction via the specific file substrate.
 */

#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <inttypes.h>
#include <stdbool.h>

#include <otf2/otf2.h>

#include <UTILS_Error.h>
#define OTF2_DEBUG_MODULE_NAME SION
#include <UTILS_Debug.h>

#include <sion.h>

#include "otf2_internal.h"
#include "otf2_file_types.h"
#include "otf2_reader.h"
#include "otf2_archive.h"

#include "OTF2_File.h"
#include "otf2_file_int.h"


/** @brief File handle to interact with sion files.
 *  Please see otf2_file_sion_struct for a detailed description. */
typedef struct otf2_file_sion_struct otf2_file_sion;

/** @brief File handle to interact with Sion files. */
struct otf2_file_sion_struct
{
    OTF2_File super;

    /* ___ Information owned by the SION file substrate. ___ */

    /** SION ID. */
    int sion_id;

    /** The rank from what we want to read. -1 if not mapped.  */
    int rank;
};

static OTF2_ErrorCode
otf2_file_sion_reset( OTF2_File* file );

static OTF2_ErrorCode
otf2_file_sion_write( OTF2_File*  file,
                      const void* buffer,
                      uint64_t    size );

static OTF2_ErrorCode
otf2_file_sion_read( OTF2_File* file,
                     void*      buffer,
                     uint64_t   size );

static OTF2_ErrorCode
otf2_file_sion_seek( OTF2_File* file,
                     int64_t    offset );

static OTF2_ErrorCode
otf2_file_sion_get_file_size( OTF2_File* file,
                              uint64_t*  size );

/** @brief Register a handle for file substrate.
 *
 *  Allocates a handle, sets the according values and registers the function
 *  pointers to its file operations.
 *
 *  Please note: The actual sion file open is executed in otf2_file_sion_init
 *  since all callbacks and parameter must be set first.
 *
 *  @param fileName         Name of the file.
 *  @param mode             Method to interact with the file.
 *  @param substrate        File substrate.
 *
 *  @return                 Returns a handle to the file if successful, NULL
 *                          otherwise.
 */
OTF2_ErrorCode
otf2_file_sion_open( int         sion_id,
                     int         rank,
                     OTF2_File** file )
{
    otf2_file_sion* sion_file =
        ( otf2_file_sion* )calloc( 1, sizeof( otf2_file_sion ) );
    if ( !sion_file )
    {
        return UTILS_ERROR( OTF2_ERROR_MEM_FAULT,
                            "Could not allocate memory for file handle!" );
    }

    sion_file->sion_id = sion_id;
    sion_file->rank    = rank;

    /* Register function pointers, for real SION files */
    sion_file->super.reset         = otf2_file_sion_reset;
    sion_file->super.write         = otf2_file_sion_write;
    sion_file->super.read          = otf2_file_sion_read;
    sion_file->super.seek          = otf2_file_sion_seek;
    sion_file->super.get_file_size = otf2_file_sion_get_file_size;

    *file = &sion_file->super;

    return OTF2_SUCCESS;
}


/** @name File operations for Sion file substrate.
 *  Please see the standard interface functions for a detailed description.
 *  @{ */
OTF2_ErrorCode
otf2_file_sion_close( OTF2_File* file )
{
    /* Get the original substrate structure. */
    otf2_file_sion* sion_file = ( otf2_file_sion* )file;

    free( sion_file );

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
otf2_file_sion_reset( OTF2_File* file )
{
    return UTILS_ERROR( OTF2_ERROR_FILE_INTERACTION,
                        "Not supported!" );
}


OTF2_ErrorCode
otf2_file_sion_write( OTF2_File*  file,
                      const void* buffer,
                      uint64_t    size )
{
    /* Get the original substrate structure. */
    otf2_file_sion* sion_file = ( otf2_file_sion* )file;

    UTILS_DEBUG( "sion_fwrite( %p, %zu, 1, %d )",
                 buffer, size, sion_file->sion_id );
    int status = sion_fwrite( buffer, size, 1, sion_file->sion_id );
    UTILS_DEBUG( "sion_fwrite( %p, %zu, 1, %d ) = %d",
                 buffer, size, sion_file->sion_id, status );

    if ( status != 1 )
    {
        return UTILS_ERROR( OTF2_ERROR_FILE_INTERACTION,
                            "Write to sion file failed: %d", status );
    }

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
otf2_file_sion_read( OTF2_File* file,
                     void*      buffer,
                     uint64_t   size )
{
    /* Get the original substrate structure. */
    otf2_file_sion* sion_file = ( otf2_file_sion* )file;

    /* Seek to get correct sion rank. */
    int status = 0;
    if ( sion_file->rank >= 0 )
    {
        UTILS_DEBUG( "sion_seek( %d, %d, SION_CURRENT_CHUNK, SION_CURRENT_POS )",
                     sion_file->sion_id,
                     sion_file->rank );
        status = sion_seek( sion_file->sion_id,
                            sion_file->rank,
                            SION_CURRENT_BLK,
                            SION_CURRENT_POS );
        UTILS_DEBUG( "sion_seek( %d, %d, SION_CURRENT_CHUNK, SION_CURRENT_POS ) = %d",
                     sion_file->sion_id,
                     sion_file->rank,
                     status );
        if ( status != 1 )
        {
            return UTILS_ERROR( OTF2_ERROR_FILE_INTERACTION,
                                "Can't seek: %d", status );
        }
    }

    UTILS_DEBUG( "sion_fread( %p, %zu, 1, %d )",
                 buffer, size, sion_file->sion_id );
    status = sion_fread( buffer, size, 1, sion_file->sion_id );
    UTILS_DEBUG( "sion_fread( %p, %zu, 1, %d ) = %d",
                 buffer, size, sion_file->sion_id, status );
    if ( status != 1 )
    {
        return UTILS_ERROR( OTF2_ERROR_FILE_INTERACTION,
                            "Read from sion file failed: %d", status );
    }

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
otf2_file_sion_seek( OTF2_File* file,
                     int64_t    offset )
{
    /* Get the original substrate structure. */
    otf2_file_sion* sion_file = ( otf2_file_sion* )file;

    int status = 0;
    UTILS_DEBUG( "sion_seek( %d, %d, SION_ABSOLUTE_POS, %d )",
                 sion_file->sion_id,
                 sion_file->rank,
                 offset );
    status = sion_seek( sion_file->sion_id,
                        sion_file->rank,
                        SION_ABSOLUTE_POS,
                        offset );
    UTILS_DEBUG( "sion_seek( %d, %d, SION_ABSOLUTE_POS, %d ) = %d",
                 sion_file->sion_id,
                 sion_file->rank,
                 offset,
                 status );

    if ( status != 1 )
    {
        return UTILS_ERROR( OTF2_ERROR_FILE_INTERACTION,
                            "Seek in sion file failed: %d", status );
    }

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
otf2_file_sion_get_file_size( OTF2_File* file,
                              uint64_t*  size )
{
    /* Get the original substrate structure. */
    otf2_file_sion* sion_file = ( otf2_file_sion* )file;

    /* Seek to get correct sion rank. */
    int status = 0;
    if ( sion_file->rank >= 0 )
    {
        UTILS_DEBUG( "sion_seek( %d, %d, SION_CURRENT_CHUNK, SION_CURRENT_POS )",
                     sion_file->sion_id,
                     sion_file->rank );
        status = sion_seek( sion_file->sion_id,
                            sion_file->rank,
                            SION_CURRENT_BLK,
                            SION_CURRENT_POS );
        UTILS_DEBUG( "sion_seek( %d, %d, SION_CURRENT_CHUNK, SION_CURRENT_POS ) = %d",
                     sion_file->sion_id,
                     sion_file->rank,
                     status );
        if ( status != 1 )
        {
            return UTILS_ERROR( OTF2_ERROR_FILE_INTERACTION,
                                "Can't seek: %d", status );
        }
    }

    int         current_chunk_num = 0;
    sion_int64  current_pos       = 0;
    int         max_chunks        = 0;
    sion_int64* chunk_sizes       = NULL;

    UTILS_DEBUG( "sion_get_current_location( %d, ... )",
                 sion_file->sion_id );
    status = sion_get_current_location( sion_file->sion_id,
                                        &current_chunk_num,
                                        &current_pos,
                                        &max_chunks,
                                        &chunk_sizes );
    UTILS_DEBUG( "sion_get_current_location( %d, ... ) = %d",
                 sion_file->sion_id, status );

    if ( status != 1 )
    {
        return UTILS_ERROR( OTF2_ERROR_FILE_INTERACTION,
                            "Get size of sion file failed: %d", status );
    }

    /* Sum up all chunk sizes to get total size. */
    uint64_t total_size = 0;
    for ( int i = 0; i < max_chunks; i++ )
    {
        total_size += chunk_sizes[ i ];
    }

    *size = total_size;

    return OTF2_SUCCESS;
}

/** @} */
